import shutil
import os
from pathlib import Path
from sqlalchemy import create_engine

from .utils.reflected import *


class DbMerge(object):
    """
    Converter DB -> feeds.
    """

    def __init__(
        self,
        input_dbs=None,
        output_db=None,
        verbose=True,
    ):
        """
        Constructor
        @param read_internet_links Read links to find RSS feeds
        @param update_feed Many things are copied from original entry.
                          If this setting is true, feed entry fetches title, and other properties
        """
        self.input_dbs = input_dbs
        self.output_db = output_db
        self.verbose = verbose

    def convert(self):
        """
        API
        """
        input_dbs = self.input_dbs
        if len(input_dbs) < 2:
            return False

        size_zero = os.path.getsize(input_dbs[0])
        size_one = os.path.getsize(input_dbs[1])

        if size_zero > size_one:
            bigger_db = input_dbs[0]
            smaller_db = input_dbs[1]
        else:
            bigger_db = input_dbs[1]
            smaller_db = input_dbs[0]

        dst = Path(self.output_db)

        if dst.exists():
            dst.unlink()

        shutil.copy(bigger_db, self.output_db)

        self.src_engine = create_engine(f"sqlite:///{smaller_db}")
        self.dst_engine = create_engine(f"sqlite:///{self.output_db}")

        with self.src_engine.connect() as connection:
            self.src_connection = connection
            with self.dst_engine.connect() as dst_connection:
                self.dst_connection = dst_connection
                self.convert_entries()

    def convert_entries(self):
        src_table = ReflectedEntryTable(self.src_engine, self.src_connection)
        for entry in src_table.get_entries_good():
            dst_table = ReflectedEntryTable(self.dst_engine, self.dst_connection)
            if not dst_table.exists(link=entry.link):
                self.convert_entry(entry)
            elif self.verbose:
                destination_entries = dst_table.get_where({"link" : entry.link})
                for destination_entry in destination_entries:
                    self.fill_blanks(entry, destination_entry, dst_table)
                print(f"Entry {entry.link} is already present")

    def convert_entry(self, entry):
        if self.verbose:
            print(f"Converting entry {entry.link}")

        copier = EntryCopier(
                             src_engine=self.src_engine,
                             src_connection=self.src_connection,
                             dst_engine = self.dst_engine,
                             dst_connection=self.dst_connection) 
        copier.copy_entry(entry)

    def fill_blanks(self, source_entry, destination_entry, destination_table):
        if (not self.is_entry_attribute_set(destination_entry, "thumbnail") and
           self.is_entry_attribute_set(source_entry, "thumbnail")):
               data = {}
               data["thumbnail"] = source_entry.thumbnail
               destination_table.update_json_data(id=destination_entry.id, json_data=data)
        if (not self.is_entry_attribute_set(destination_entry, "title") and
           self.is_entry_attribute_set(source_entry, "title")):
               data = {}
               data["title"] = source_entry.title
               destination_table.update_json_data(id=destination_entry.id, json_data=data)
        if (not self.is_entry_attribute_set(destination_entry, "description") and
           self.is_entry_attribute_set(source_entry, "description")):
               data = {}
               data["description"] = source_entry.description
               destination_table.update_json_data(id=destination_entry.id, json_data=data)

    def is_entry_attribute_set(self, entry, attribute):
        attribute = getattr(entry, attribute, None)
        if not attribute:
            return False

        return True



"""
@brief Converts JSON files to SQLite DB

SQLite can easily be imported and used by other projects.
"""

import os
import sqlite3
import json
import argparse
import time
from pathlib import Path
from sqlalchemy import create_engine
from dateutil import parser

from .utils.reflected import *


class DirReader(object):
    def __init__(self, source_files_directory, accepted_extensions=None):
        self.dir = source_files_directory
        if accepted_extensions is None:
            self.accepted_extensions = [".json"]

    def get_files(self):
        file_list = []
        for root, dirs, files in os.walk(self.dir):
            for file in files:
                file_split = os.path.splitext(file)
                if file_split[1] in self.accepted_extensions:
                    file_list.append(os.path.join(root, file))

        file_list = sorted(file_list)
        return file_list


class JSON2Db(object):
    """
    Performs actual conversion between from JSON to DB
    """

    def __init__(self, input_file=None, input_dir=None, output_db=None, preserve_id=False, vote_threshold=None, verbose=False):
        self.input_file = input_file
        self.input_dir = input_dir
        self.output_db = output_db
        self.preserve_id = preserve_id
        self.vote_threshold = vote_threshold
        self.verbose = verbose

        if self.input_dir:
            self.file_reader = DirReader(source_files_directory=self.input_dir)
            self.files = self.file_reader.get_files()
        elif self.input_file:
            self.file_reader = None
            self.files = [self.input_file]
        else:
            self.file_reader = None
            self.files = []

    def convert(self):
        #path = Path(self.output_db)
        #if path.exists():
        #    path.unlink()

        self.engine = create_engine(f"sqlite:///{self.output_db}")
        with self.engine.connect() as connection:
            self.connection = connection

            total_num_files = len(self.files)

            for row, afile in enumerate(self.files):
                print("[{}/{}]: file:{}".format(row, total_num_files, afile))
                self.convert_file(afile)

    def convert_file(self, file_name):
        data = self.read_file(file_name)
        if not data:
            return

        total_rows = len(data)

        for row, entry in enumerate(data):
            entry = self.prepare_entry(entry)
            if "link" in entry:
                if self.preserve_id:
                    if "id" not in entry:
                        print("Entry {} is missing ID".format(entry["link"]))
                        continue
                else:
                    entry["id"] = row

                if self.is_entry_to_be_added(entry):
                    table = ReflectedEntryTable(engine=self.engine, connection=self.connection)
                    if table.insert_json(entry) is not None:
                        if self.verbose:
                            print(
                                " -> [{}/{}] Link:{} Added".format(
                                    row, total_rows, entry["link"]
                                )
                            )
                    else:
                        print(
                            " -> [{}/{}] Link:{} NOT Added".format(
                                row, total_rows, entry["link"]
                            )
                        )
                else:
                    if self.verbose:
                        print(
                            " -> [{}/{}] Link:{} Skipped".format(
                                row, total_rows, entry["link"]
                            )
                        )

    def prepare_entry(self, entry):
        """
        Drops any unwelcome keys
        """
        table = ReflectedEntryTable(engine=self.engine, connection=self.connection)
        columns = table.get_column_names()
        keys = list(entry.keys())

        diff = list(set(keys) - set(columns))
        for item in diff:
            del entry[item]

        for key in entry:
            if key.startswith("date"):
                if entry[key]:
                    entry[key] = parser.parse(entry[key])

        return entry

    def is_entry_to_be_added(self, entry):
        # entry already exists
        table = ReflectedEntryTable(engine=self.engine, connection=self.connection)
        if "id" in entry and table.exists(id=entry["id"]):
            return False
        if "link" in entry and table.exists(link=entry["link"]):
            return False

        if self.vote_threshold:
            if "page_rating_votes" in entry:
                if entry["page_rating_votes"]:
                    if int(entry["page_rating_votes"]) < self.vote_threshold:
                        return False
                    else:
                        return True
                return False
            return False

        return True

    def read_file_contents(self, file_name):
        with open(file_name, "r") as f:
            return f.read()

    def read_file(self, file_name):
        text = self.read_file_contents(file_name)

        try:
            j = json.loads(text)

            if "links" in j:
                return j["links"]
            if "sources" in j:
                return j["sources"]

            return j
        except Exception as e:
            print("Could not read file: {}".format(afile))


class Parser(object):
    def parse(self):
        self.parser = argparse.ArgumentParser(description="Data converter program")
        self.parser.add_argument("--input-file", help="File to be scanned")
        self.parser.add_argument("--input-dir", help="Directory to be scanned")
        self.parser.add_argument(
            "--output-db", default="converted.sqlite", help="Output db name"
        )
        self.parser.add_argument(
            "--preserve-id", action="store_true", help="Preserves ID of objects"
        )
        self.parser.add_argument("--vote-min", help="Minimum amount of entry vote")
        self.parser.add_argument("--language", help="Accept language")  # TODO implement
        self.parser.add_argument("--entries", help="Convert entries")  # TODO implement
        self.parser.add_argument("--sources", help="Convert sources")  # TODO implement
        self.parser.add_argument(
            "--verbose", action="store_true", help="Shows more info"
        )

        self.args = self.parser.parse_args()

        if self.args.dir:
            self.dir = self.args.dir
        else:
            self.dir = None

        if self.args.preserve_id:
            self.preserve_id = self.args.preserve_id
        else:
            self.preserve_id = None

        if self.args.vote_min:
            self.vote_min = int(self.args.vote_min)
        else:
            self.vote_min = None


def main():
    print("Starting processing")
    parser = Parser()
    parser.parse()

    try:
        start_time = time.time()

        c = JSON2Db(input_file = parser.args.input_file, input_dir=parser.args.input_dir, output_db = parser.args.output_db)
        c.convert()

        elapsed_time_seconds = time.time() - start_time
        elapsed_minutes = int(elapsed_time_seconds // 60)
        elapsed_seconds = int(elapsed_time_seconds % 60)
        print(f"Time: {elapsed_minutes}:{elapsed_seconds}")

    except Exception as e:
        print("Exception: {}".format(e))
    except KeyboardInterrupt as e:
        print("Exception: {}".format(e))

    db.close()
    print("Processing DONE")


if __name__ == "__main__":
    main()



"""
Converts Database to information about RSS feeds.

TODO - copy tags from origin to RSS feeds
"""

import shutil
import argparse
from pathlib import Path
from sqlalchemy import create_engine

from webtoolkit import RemoteUrl, BaseUrl
from linkarchivetools import tableconfig
from .utils.reflected import *


class Db2Feeds(object):
    """
    Converter DB -> feeds.
    """

    def __init__(
        self,
        input_db=None,
        output_db=None,
        verbose=True,
        clean=False,
        remote_server="",
        output_format=None,
        read_internet_links=False,
        update_feed=False,
    ):
        """
        Constructor
        @param read_internet_links Read links to find RSS feeds
        @param update_feed Many things are copied from original entry.
                          If this setting is true, feed entry fetches title, and other properties
        """
        self.input_db = input_db
        self.output_db = output_db
        self.verbose = verbose
        self.clean = clean
        self.remote_server = remote_server
        self.output_format = output_format
        self.read_internet_links = read_internet_links
        self.update_feed = update_feed

        self.new_table = None

        if self.output_db:
            self.output_format = "SQLITE"
        self.make_output_db()

    def make_output_db(self):
        if self.output_format != "SQLITE":
            self.new_engine = None
            self.new_connection = None
            return

        new_path = Path(self.output_db)
        if new_path.exists() and self.clean:
            new_path.unlink()

        if not new_path.exists():
            shutil.copy(self.input_db, self.output_db)
            self.new_engine = create_engine(f"sqlite:///{self.output_db}")
            with self.new_engine.connect() as new_connection:
                self.new_connection = new_connection
                self.truncate_tables()

        self.new_engine = create_engine(f"sqlite:///{self.output_db}")

    def convert(self):
        """
        API
        """
        self.engine = create_engine(f"sqlite:///{self.input_db}")
        with self.engine.connect() as connection:
            self.connection = connection
            if self.new_engine:
                with self.new_engine.connect() as new_connection:
                    self.new_connection = new_connection
                    self.convert_entries()
            else:
                self.convert_entries()

    def convert_entries(self):
        if self.clean:
            self.truncate_tables()

        self.new_table = None
        if self.new_engine:
            self.new_table = ReflectedEntryTable(self.new_engine, self.new_connection)

        table = ReflectedEntryTable(self.engine, self.connection)
        for entry in table.get_entries_good():
            self.convert_entry(entry)

    def convert_entry(self, entry):
        url = BaseUrl(entry.link)
        feeds = url.get_feeds()

        if len(feeds) == 0:
            if self.read_internet_links:
                if self.remote_server:
                    url_ex = RemoteUrl(remote_server_location=self.remote_server, url=entry.link)
                    url_ex.get_response()
                    feeds.extend(url_ex.get_feeds())

        for feed in feeds:
            data = self.prepare_data(entry, feed)

            if self.new_table:
                if not self.new_table.exists(link=feed):
                    self.print_data(entry, data)
                    self.copy_entry(entry, self.new_table, data)
            else:
                self.print_data(entry, data)

    def prepare_data(self, entry, feed):
        data = {}
        data["link"] = feed
        data["title"] = entry.title
        data["page_rating_votes"] = entry.page_rating_votes
        data["manual_status_code"] = entry.manual_status_code
        data["thumbnail"] = entry.thumbnail
        data["language"] = entry.language

        # not null requirement
        data["source_url"] = ""
        data["permanent"] = False
        data["bookmarked"] = False
        data["status_code"] = entry.status_code
        data["contents_type"] = 0
        data["page_rating_contents"] = 0
        data["page_rating_visits"] = 0
        data["page_rating"] = 0

        if self.update_feed and self.remote_server:
            url_feed = RemoteUrl(remote_server=self.remote_server, url=feed)
            url_feed.get_response()

            data["title"] = url_feed.get_title()
            data["description"] = url_feed.get_description()
            data["status_code"] = url_feed.get_status_code()
            data["thumbnail"] = url_feed.get_thumbnail()
        return data

    def copy_entry(self, entry, entry_table, data):
        """
        """
        new_entry_id = entry_table.insert_json(data)
        self.copy_tags(entry, new_entry_id)
        self.copy_social_data(entry, new_entry_id)

    def copy_tags(self, entry, new_entry_id):
        source_entry_compacted_tags = ReflectedEntryCompactedTags(self.engine, self.connection)
        tags = source_entry_compacted_tags.get_tags(entry.id)

        entry_tag_data = {}
        for tag in tags:
            entry_tag_data["tag"] = tag
            entry_tag_data["entry_id"] = new_entry_id
            destination_entry_compacted_tags = ReflectedEntryCompactedTags(self.new_engine, self.new_connection)
            destination_entry_compacted_tags.insert_json_data(entry_tag_data)

    def copy_social_data(self, entry, new_entry_id):
        source_entry_social_data = ReflectedSocialData(self.engine, self.connection)
        social_data = source_entry_social_data.get_json(entry.id)
        if social_data:
            if "id" in social_data:
                del social_data["id"]
            social_data["entry_id"] = new_entry_id

            destination_entry_social_data = ReflectedSocialData(self.new_engine, self.new_connection)
            destination_entry_social_data.insert_json_data(social_data)

    def truncate_tables(self):
        if not self.new_engine:
            return

        table_names = self.get_table_names()
        for table_name in table_names:
            table = ReflectedTable(self.new_engine, self.new_connection)
            table.truncate_table(table_name)

        table = ReflectedTable(self.new_engine, self.new_connection)
        table.vacuum()

    def get_table_names(self):
        return tableconfig.get_backup_tables()

    def print_data(self, entry, data):
        """
        If we print to SQLITE we want to see progress so we display it anyway
        """
        if not self.verbose:
            return

        link = data["link"]
        title = data["title"]
        page_rating_votes = data["page_rating_votes"]

        if self.output_format == "LINES" or self.output_format == "SQLITE":
            print(f"[{page_rating_votes}] {link} - {title}")
            user_tags = ReflectedEntryCompactedTags(self.engine, self.connection)
            tags = user_tags.get_tags_string(entry.id)
            if tags:
                print(f"{tags}")
        elif self.output_format == "JSON":
            user_tags = ReflectedEntryCompactedTags(self.engine, self.connection)
            tags = user_tags.get_tags(entry.id)
            print(
                f"""
            \{ "title" : "{title}",
              "link" : "{link}",
              "page_rating_votes : {page_rating_votes},
              "tags" : {tags}
            \}"""
            )
        else:
            print("Unsupported output format")


def parse():
    parser = argparse.ArgumentParser(description="Data analyzer program")
    parser.add_argument("--db", default="catalog.db", help="DB to be scanned")
    parser.add_argument("--output-db", help="File to be created")
    parser.add_argument("--update-rss",action="store_true", help="Reads RSS to check it's title and properties")
    parser.add_argument("--clean",action="store_true", help="If output db exists, then it is removed at start")
    parser.add_argument("--read-internet-links",action="store_true", help="Reads entries to check if contains RSS. Without it only calculated RSS are returned")
    parser.add_argument(
        "--output-format",
        default="LINES",
        help="format of display. LINES, JSON, SQLITE",
    )
    parser.add_argument("--crawling-server", default="", help="Remote crawling server")

    args = parser.parse_args()

    return parser, args


def main():
    parser, args = parse()

    path = Path(args.db)
    if not path.exists():
        print("File {} does not exist".format(path))
        return

    reader = Db2Feeds(
        input_db=args.db,
        output_db=args.output_db,
        clean=args.clean,
        remote_server=args.crawling_server,
        output_format=args.output_format,
    )
    reader.convert()


if __name__ == "__main__":
    main()




from .backup import Backup, parse_backup_commandline
from .db2feeds import Db2Feeds
from .dbmerge import DbMerge
from .db2json import Db2JSON
from .dbanalyzer import DbAnalyzer
from .dbfilter import DbFilter
from .json2db import JSON2Db



"""
This file is mainly for SQLite.
It will not open several connection, will use one.
This allows us to handle nested calls of generators without any problems.
"""
from sqlalchemy import (
    MetaData,
    Table,
    select,
    delete,
    or_,
    and_,
    exists,
    text,
    inspect,
    insert,
    update,
    Index,
)


class ReflectedTable(object):
    def __init__(self, engine, connection):
        self.engine = engine
        self.connection = connection

    def get_table(self, table_name):
        destination_metadata = MetaData()
        destination_table = Table(
            table_name, destination_metadata, autoload_with=self.engine
        )
        return destination_table

    def truncate_table(self, table_name):
        sql_text = f"DELETE FROM {table_name};"
        self.connection.execute(text(sql_text))
        self.connection.commit()

    def create_index(self, table, column_name):
        index_name = f"idx_{table.name}_{column_name}"
        index = Index(index_name, getattr(table.c, column_name))

        index.create(bind=self.engine)

    def enable_sqlite_wal(self, table_name):
        sql_text = f"PRAGMA journal_mode=WAL;"
        self.connection.execute(text(sql_text))
        self.connection.commit()

    def vacuum(self):
        self.connection.execute(text("VACUUM"))

    def close(self):
        pass

    def insert_json_data(self, table_name, json_data: dict):
        table = self.get_table(table_name)

        stmt = (
            insert(table)
            .values(**json_data)
            .returning(table.c.id)
        )

        result = self.connection.execute(stmt)
        inserted_id = result.scalar_one()
        self.connection.commit()

        return inserted_id

    def count(self, table_name):
        row_count = self.connection.execute(
            text(f"SELECT COUNT(*) FROM {table_name}")
        ).scalar()
        return row_count

    def print_summary(self, print_columns=False):
        tables = self.get_table_names()

        for table in tables:
            row_count = self.count(table)
            print(f"Table: {table}, Row count: {row_count}")

            if print_columns:
                column_names = self.get_column_names(table)
                print(f"Columns in {table}: {', '.join(column_names)}")

    def get_table_names(self):
        inspector = inspect(self.engine)
        tables = inspector.get_table_names()
        return list(tables)

    def get_column_names(self, table):
        inspector = inspect(self.engine)

        columns = inspector.get_columns(table)
        column_names = [column["name"] for column in columns]
        return column_names

    def row_to_json_data(self, row):
        data = dict(row._mapping)
        return data

    def run_sql(self, sql_text):
        self.connection.execute(text(sql_text))
        self.connection.commit()


class ReflectedGenericTable(object):
    def __init__(self, engine, connection, table_name=None):
        self.engine = engine
        self.connection = connection
        self.table_name = table_name
        self.table = None

        if self.table_name is None:
            self.table_name = self.get_table_name()

    def get_table_name():
        return self.table_name

    def get_table(self):
        if self.table is not None:
            return self.table

        destination_metadata = MetaData()
        self.table = Table(
            self.table_name, destination_metadata, autoload_with=self.engine
        )
        return self.table

    def truncate(self):
        sql_text = f"DELETE FROM {self.table_name};"
        self.connection.execute(text(sql_text))
        self.connection.commit()

    def enable_sqlite_wal(self):
        sql_text = f"PRAGMA journal_mode=WAL;"
        self.connection.execute(text(sql_text))
        self.connection.commit()

    def create_index(self, column_name):
        index_name = f"idx_{self.table.name}_{column_name}"
        index = Index(index_name, getattr(table.c, column_name))

        index.create(bind=self.engine)

    def insert_json_data(self, json_data: dict):
        table = self.get_table()

        stmt = (
            insert(table)
            .values(**json_data)
            .returning(table.c.id)
        )

        result = self.connection.execute(stmt)
        inserted_id = result.scalar_one()
        self.connection.commit()

        return inserted_id

    def update_json_data(self, id, json_data):
        table = self.get_table()

        stmt = (
            update(table)
            .where(table.c.id == id)
            .values(**json_data)
        )

        self.connection.execute(stmt)
        self.connection.commit()

    def count(self):
        row_count = self.connection.execute(
            text(f"SELECT COUNT(*) FROM {self.table_name}")
        ).scalar()
        return row_count

    def get(self, id):
        destination_table = self.get_table()

        stmt = select(destination_table).where(destination_table.c.id == id)

        result = self.connection.execute(stmt)
        return result.first()

    def get_where(self,
                  conditions_map: dict=None,
                  conditions=None,
                  order_by=None,
                  limit:int|None=None,
                  offset:int=0):
        """
        @param conditions_map can be passed as {"name": "Test"}
        @param conditions can be passed as [destionation_table.c.rating > 5]
        @param order_by can be passed as [destionation_table.c.name.asc()]
        """
        destination_table = self.get_table()

        if not conditions:
            conditions = []

        if not conditions and conditions_map:
            for column_name, value in conditions_map.items():
                if not hasattr(destination_table.c, column_name):
                    raise ValueError(f"Unknown column: {column_name}")

                conditions.append(getattr(destination_table.c, column_name) == value)

        stmt = select(destination_table)

        if conditions:
            stmt = stmt.where(or_(*conditions))
        if order_by:
            stmt = stmt.order_by(*order_by)
        if offset:
            stmt = stmt.offset(offset)
        if limit is not None:
            stmt = stmt.limit(limit)

        result = self.connection.execute(stmt)
        for row in result:
            yield row

    def delete(self, id):
        destination_table = self.get_table()

        stmt = delete(destination_table).where(destination_table.c.id == id)

        result = self.connection.execute(stmt)
        self.connection.commit()

        return result.rowcount  # number of rows deleted

    def delete_where(self, conditions: dict):
        destination_table = self.get_table()

        filters = []
        for column_name, value in conditions.items():
            if not hasattr(destination_table.c, column_name):
                raise ValueError(f"Unknown column: {column_name}")

            filters.append(getattr(destination_table.c, column_name) == value)

        stmt = delete(destination_table).where(and_(*filters))

        result = self.connection.execute(stmt)
        self.connection.commit()

        return result.rowcount

    def print_summary(self, print_columns=False):
        row_count = self.count()
        print(f"Table: {self.table_name}, Row count: {row_count}")

        if print_columns:
            column_names = self.get_column_names()
            print(f"Columns in {self.table_name}: {', '.join(column_names)}")

    def get_column_names(self):
        inspector = inspect(self.engine)
        row_count = self.connection.execute(
            text(f"SELECT COUNT(*) FROM {self.table_name}")
        ).scalar()

        columns = inspector.get_columns(self.table_name)
        column_names = [column["name"] for column in columns]
        return column_names

    def row_to_json_data(self, row):
        data = dict(row._mapping)
        return data

    def run_sql(self, sql_text):
        self.connection.execute(text(sql_text))
        self.connection.commit()


class ReflectedEntryTable(ReflectedGenericTable):
    def get_table_name(self):
        return "linkdatamodel"

    def insert_json(self, entry_json):
        if "link" not in entry_json:
            return

        if "source_url" not in entry_json:
            entry_json["source_url"] = ""
        if "permanent" not in entry_json:
            entry_json["permanent"] = False
        if "bookmarked" not in entry_json:
            entry_json["bookmarked"] = False
        if "status_code" not in entry_json:
            entry_json["status_code"] = 0
        if "contents_type" not in entry_json:
            entry_json["contents_type"] = 0
        if "page_rating_contents" not in entry_json:
            entry_json["page_rating_contents"] = 0
        if "page_rating_visits" not in entry_json:
            entry_json["page_rating_visits"] = 0
        if "page_rating_votes" not in entry_json:
            entry_json["page_rating_votes"] = 0
        if "page_rating" not in entry_json:
            entry_json["page_rating"] = 0

        return self.insert_json_data(entry_json)

    def get_entries(self, limit:int|None=None, offset:int=0):
        destination_table = self.get_table()

        entries_select = select(destination_table)

        if offset:
            entries_select = entries_select.offset(offset)
        if limit is not None:
            entries_select = entries_select.limit(limit)

        result = self.connection.execute(entries_select)

        for entry in result:
            yield entry

    def get_entries_good(self):
        destination_table = self.get_table()

        stmt = (
            select(destination_table)
            .where(destination_table.c.page_rating_votes > 0)
            .order_by(destination_table.c.page_rating_votes.desc())
        )

        result = self.connection.execute(stmt)
        entries = result.fetchall()

        for entry in entries:
            yield entry

    def exists(self, *, id=None, link=None):
        table = self.get_table()

        conditions = []
        if id is not None:
            conditions.append(table.c.id == id)
        if link is not None:
            conditions.append(table.c.link == link)

        if not conditions:
            return False

        stmt = select(exists().where(or_(*conditions)))
        return self.connection.execute(stmt).scalar()


class ReflectedUserTags(ReflectedGenericTable):
    def get_table_name(self):
        return "usertags"

    def get_tags_string(self, entry_id):
        destination_table = self.get_table()

        stmt = select(destination_table).where(destination_table.c.entry_id == entry_id)

        tags = ""

        result = self.connection.execute(stmt)
        rows = result.fetchall()
        for row in rows:
            if tags:
                tags += ", "

            tags += "#" + row.tag

        return tags

    def get_tags(self, entry_id):
        destination_table = self.get_table()

        stmt = select(destination_table).where(destination_table.c.entry_id == entry_id)

        tags = []

        result = self.connection.execute(stmt)
        rows = result.fetchall()
        for row in rows:
            tags.append(row.tag)

        return tags


class ReflectedEntryCompactedTags(ReflectedGenericTable):
    def get_table_name(self):
        return "entrycompactedtags"

    def get_tags_string(self, entry_id):
        destination_table = self.get_table()

        stmt = select(destination_table).where(destination_table.c.entry_id == entry_id)

        tags = ""

        result = self.connection.execute(stmt)
        rows = result.fetchall()
        for row in rows:
            if tags:
                tags += ", "

            tags += "#" + row.tag

        return tags

    def get_tags(self, entry_id):
        destination_table = self.get_table()

        stmt = select(destination_table).where(destination_table.c.entry_id == entry_id)

        tags = []

        result = self.connection.execute(stmt)
        rows = result.fetchall()
        for row in rows:
            tags.append(row.tag)

        return tags


class ReflectedSourceTable(ReflectedGenericTable):
    def get_table_name(self):
        return "sourcedatamodel"

    def get_source(self, source_id):
        destination_table = self.get_table()

        stmt = select(destination_table).where(destination_table.c.id == source_id)

        result = self.connection.execute(stmt)
        return result.first()

    def get_sources(self, limit:int|None=None, offset:int=0):
        destination_table = self.get_table()

        sources_select = select(destination_table)

        if offset:
            sources_select = sources_select.offset(offset)
        if limit is not None:
            sources_select = sources_select.limit(limit)

        result = self.connection.execute(sources_select)

        for source in result:
            yield source

    def insert_json(self, source_json):
        if "url" not in source_json:
            source_json["url"] = ""

        return self.insert_json_data(source_json)

    def exists(self, *, id=None, url=None):
        table = self.get_table()

        conditions = []
        if id is not None:
            conditions.append(table.c.id == id)
        if url is not None:
            conditions.append(table.c.url == url)

        if not conditions:
            return False

        stmt = select(exists().where(or_(*conditions)))
        return self.connection.execute(stmt).scalar()


class ReflectedSourceOperationalData(ReflectedGenericTable):
    def get_table_name(self):
        return "sourceoperationaldata"

    def get_for_source(self, source_id):
        destination_table = self.get_table()

        stmt = select(destination_table).where(destination_table.c.source_obj_id == source_id)

        result = self.connection.execute(stmt)
        return result.first()


class ReflectedSocialData(ReflectedGenericTable):
    def get_table_name(self):
        return "socialdata"

    def get(self, entry_id):
        destination_table = self.get_table()

        stmt = select(destination_table).where(destination_table.c.entry_id == entry_id)

        result = self.connection.execute(stmt)
        return result.first()

    def get_json(self, entry_id):
        row = self.get(entry_id)
        if row is None:
            return None

        data = self.row_to_json_data(row)
        return data


class ReflectedEntryRules(ReflectedGenericTable):
    def get_table_name(self):
        return "entryrules"


class ReflectedConfigurationEntry(ReflectedGenericTable):
    def get_table_name(self):
        return "configurationentry"

    def get(self):
        first_item = self.get_first()
        if first_item:
            return first_item

        self.add_configuration()

        first_item = self.get_first()
        if first_item:
            return first_item

    def get_first(self):
        destination_table = self.get_table()

        stmt = select(destination_table)

        result = self.connection.execute(stmt)
        return result.first()

    def add_configuration(self):
        json_data = {}
        self.insert_json_data(json_data)


class ReflectedAppLogging(ReflectedGenericTable):
    def get_table_name(self):
        return "applogging"


class EntryCopier(object):
    def __init__(self, src_engine, src_connection, dst_engine, dst_connection):
        self.src_engine = src_engine
        self.src_connection = src_connection

        self.dst_engine = dst_engine
        self.dst_connection = dst_connection

    def copy_entry(self, entry):
        """
        """
        entry_table = ReflectedEntryTable(self.dst_engine, self.dst_connection)
        data = entry_table.row_to_json_data(entry)
        del data["id"]
        new_entry_id = entry_table.insert_json(data)
        if new_entry_id is not None:
            self.copy_tags(entry, new_entry_id)
            self.copy_social_data(entry, new_entry_id)
        return new_entry_id

    def copy_tags(self, entry, new_entry_id):
        source_entry_compacted_tags = ReflectedEntryCompactedTags(self.src_engine, self.src_connection)
        tags = source_entry_compacted_tags.get_tags(entry.id)

        entry_tag_data = {}
        for tag in tags:
            entry_tag_data["tag"] = tag
            entry_tag_data["entry_id"] = new_entry_id
            destination_entry_compacted_tags = ReflectedEntryCompactedTags(self.dst_engine, self.dst_connection)
            destination_entry_compacted_tags.insert_json_data(entry_tag_data)

    def copy_social_data(self, entry, new_entry_id):
        source_entry_social_data = ReflectedSocialData(self.src_engine, self.src_connection)
        social_data = source_entry_social_data.get_json(entry.id)
        if social_data:
            if "id" in social_data:
                del social_data["id"]
            social_data["entry_id"] = new_entry_id

            destination_entry_social_data = ReflectedSocialData(self.dst_engine, self.dst_connection)
            destination_entry_social_data.insert_json_data(social_data)



"""
POSTGRES backup script.

 - Creates and restores backups (custom format, SQLITE format)
 - Check against data corruption with analyze, vacuum and reindex

What to do:
 - run vacuum, which shows problems with data
 - if any table is affected run reindex on it
 - run reindex from time to time
"""
import sys
import os
import subprocess
import argparse
from pathlib import Path
import time
from datetime import datetime

from sqlalchemy import create_engine, Column, String, Integer, MetaData, Table, text, LargeBinary, DateTime, select
from sqlalchemy.dialects.postgresql.types import BYTEA
from sqlalchemy.orm import sessionmaker

from linkarchivetools.utils.reflected import *
from linkarchivetools.tableconfig import get_backup_tables, get_tables


output_directory = Path(__file__).parents[1]


tables_to_create = get_tables()
tables_to_backup = get_backup_tables()


all_tables = list(tables_to_backup)
all_tables.append("blockentry")
all_tables.append("apikeys")
all_tables.append("applogging")
all_tables.append("backgroundjob")
all_tables.append("backgroundjobhistory")
all_tables.append("keywords")

def get_backup_directory(export_type):
    return output_directory / "data" / ("backup_" + export_type)


def get_workspace_backup_directory(export_type, workspace):
    return get_backup_directory(export_type) / workspace


def run_pg_dump_backup(run_info):
    workspace = run_info["workspace"]
    tables = run_info["tables"]
    output_file = run_info["output_file"]
    user = run_info["user"]
    database = run_info["database"]
    host = run_info["host"]

    if "format" not in run_info:
        run_info["format"] = "custom"

    if run_info["format"] == "custom":
        format_args = "c"
    elif run_info["format"] == "plain" or run_info["format"] == "sql":
        format_args = "p"

    command_input = [
        "pg_dump",
        "-h", host,
        "-U", user,
        "-d", database,
        "-F", format_args,
        "-f", output_file,
        "--data-only",
    ]

    if "format" in run_info and run_info["format"] == "sql":
        command_input.append("--inserts")

    for table in tables:
        command_input.append("-t")
        command_input.append(table)

    operating_dir = get_workspace_backup_directory(run_info["format"], workspace)
    operating_dir.mkdir(parents=True, exist_ok=True)

    print("Running: {} @ {}".format(command_input, operating_dir))

    try:
        result = subprocess.run(command_input, cwd=str(operating_dir), check=True, 
                                stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
        print("Backup completed successfully.")
    except subprocess.CalledProcessError as e:
        print("An error occurred:", e)
        print("Standard Output:", e.stdout)
        print("Standard Error:", e.stderr)
        return False

    return True


def truncate_table(run_info, table):
    user = run_info["user"]
    database = run_info["database"]
    host = run_info["host"]
    tables = run_info["tables"]

    print("Truncating {}".format(table))

    sql = f'TRUNCATE TABLE {table} CASCADE;'

    command = [
       'psql',
       "-h", host,
       "-U", user,
       "-d", database,
       '-c', sql,
    ]

    try:
        subprocess.run(command, check=True)
        print("Table truncated successfully.")
    except subprocess.CalledProcessError as e:
        print("An error occurred:", e)
        return False

    return True


def reset_table_index_sequence(run_info, table):
    user = run_info["user"]
    database = run_info["database"]
    host = run_info["host"]
    tables = run_info["tables"]

    print("Resetting sequence for table {}".format(table))

    sql = f"SELECT setval('{table}_id_seq', COALESCE((SELECT MAX(id) FROM {table}), 1));"

    command = [
       'psql',
       "-h", host,
       "-U", user,
       "-d", database,
       '-c', sql,
    ]

    try:
        subprocess.run(command, check=True)
        print("Reset index sequence successfully.")
    except subprocess.CalledProcessError as e:
        print("An error occurred:", e)
        return False

    return True


def reset_tables_index_sequence(tablemapping, run_info):
    workspace = run_info["workspace"]

    # after restore we need to reset sequences
    for item in tablemapping:
        key = item[0]
        tables = item[1]

        run_info["output_file"] = key
        run_info["tables"] = tables

        for table in tables:
            table = table.replace("instance_", workspace+"_")

            if not reset_table_index_sequence(run_info, table):
                print("Could not reset index in table")
                return False

    return True


def run_table_sql(run_info, table, sql):
    user = run_info["user"]
    database = run_info["database"]
    host = run_info["host"]

    print("Call {} initiating".format(table))

    command = [
       'psql',
       "-h", host,
       "-U", user,
       "-d", database,
       '-c', sql,
    ]

    try:
        subprocess.run(command, check=True)
        print("Call {} successful.".format(sql))
    except subprocess.CalledProcessError as e:
        print("An error occurred:", e)
        return False

    return True


def truncate_all(run_info):
    workspace = run_info["workspace"]
    user = run_info["user"]
    database = run_info["database"]
    host = run_info["host"]
    tables = run_info["tables"]

    for table in tables:
        table = table.replace("instance_", workspace+"_")

        if not truncate_table(run_info, table):
            return False

    return True


def run_pg_restore(run_info):
    workspace = run_info["workspace"]
    tables = run_info["tables"]
    output_file = run_info["output_file"]
    user = run_info["user"]
    database = run_info["database"]
    host = run_info["host"]

    if "format" not in run_info:
        run_info["format"] = "custom"

    if run_info["format"] == "custom":
        format_args = "c"
    elif run_info["format"] == "plain":
        format_args = "p"

    command_input = [
        "pg_restore",
        "-h", host,
        "-U", user,
        "-d", database,
        "-F", format_args,
        "--data-only",
        output_file,
    ]

    for table in tables:
        command_input.append("-t")
        command_input.append(table)

    operating_dir = get_workspace_backup_directory(run_info["format"], workspace)
    operating_dir.mkdir(parents=True, exist_ok=True)

    print("Running: {} @ {}".format(command_input, operating_dir))

    try:
        subprocess.run(command_input, cwd = str(operating_dir), check=True)
        print("Restore completed successfully.")
    except subprocess.CalledProcessError as e:
        print("An error occurred:", e)
        return False

    return True


### SQLite code

def create_destionation_table(table_name, source_table, destination_engine):
    """
    Copy columns from postgres to sqlite
    BYTEA is not represented in sqlite
    """
    with destination_engine.connect() as connection:
        if not destination_engine.dialect.has_table(connection, table_name):
            columns = []
            for column in source_table.columns:
                col_type = column.type
                if isinstance(col_type, BYTEA):
                    col_type = LargeBinary

                # Handle 'id' as primary key and autoincrement in SQLite
                if column.name == "id":
                    columns.append(Column(
                        "id",
                        Integer,
                        primary_key=True,
                        autoincrement=True,
                        nullable=False
                    ))
                else:
                    columns.append(Column(
                        column.name,
                        col_type,
                        nullable=column.nullable
                    ))

                # For debugging purposes, you can print column details
                # print(column.name)
                # print(column.type.__class__)
                # print(f"Nullable: {column.nullable}")

            destination_metadata = MetaData()
            destination_table = Table(table_name, destination_metadata, *columns)
            destination_table.create(destination_engine)


def get_engine_table(workspace, table_name, engine, with_workspace=True):
    if with_workspace:
        this_tables_name = "{}_{}".format(workspace, table_name)
    else:
        this_tables_name = table_name

    engine_metadata = MetaData()
    engine_table = Table(this_tables_name, engine_metadata, autoload_with=engine)
    return engine_table


def get_table_row_values(row, source_table):
    data = {}
    
    for column in source_table.columns:
        value = getattr(row, column.name)
        data[column.name] = value

    return data


def is_row_with_id(connection, table, id_value):
    try:
        existing = connection.execute(
            select(table.c.id).where(table.c.id == id_value)
        ).first()

        return existing
    except Exception as E:
        connection.rollback()
        return False


def copy_table(instance, table_name, source_engine, destination_engine, override=False, to_sqlite=True, commit_every_row=False):
    """
    Copies table from postgres to destination
    @param override If true, will work faster, since it does not check if row with id exists
    @param to_sqlite If to SQLlite then destination table names will not include workspace. If from SQLite then
                  source tables will not include
    """
    Session = sessionmaker(bind=source_engine)
    session = Session()

    source_table = get_engine_table(instance, table_name, source_engine, with_workspace=to_sqlite)
    destination_table = get_engine_table(instance, table_name, destination_engine, with_workspace=not to_sqlite)

    print(f"Copying from {source_table} to {destination_table}")

    with destination_engine.connect() as destination_connection:
        with source_engine.connect() as connection:
            result = connection.execute(source_table.select())

            index = 0
            for row in result:
                index += 1
                sys.stdout.write("{}\r".format(index))

                data = get_table_row_values(row, source_table)

                # Check if the ID already exists
                if not override:
                    id_value = data.get("id")
                    if id_value is not None:
                        existing = is_row_with_id(destination_connection, destination_table, id_value)

                        if existing:
                            continue
                
                try:
                   destination_connection.execute(destination_table.insert(), data)
                except Exception as e:
                    print(f"Skipping row {index} due to insert error {e}")
                    destination_connection.rollback()
                    continue

                if commit_every_row:
                    try:
                        destination_connection.commit()
                    except Exception as e:
                        print(f"Skipping row {index} due to insert error {e}")
                        destination_connection.rollback()
                        continue

            if not commit_every_row:
                destination_connection.commit()

    session.close()


def obfuscate_user_table(table_name, destination_engine):
    """
    Remove passwords from the database
    """
    destination_metadata = MetaData()
    destination_table = Table(table_name, destination_metadata, autoload_with=destination_engine)

    columns = destination_table.columns.keys()
    is_superuser_index = columns.index('is_superuser')

    with destination_engine.connect() as destination_connection:
        result = destination_connection.execute(destination_table.select())

        for row in result:
            update_stmt = destination_table.update().where(destination_table.c.id == row[0]).values(password='')
            destination_connection.execute(update_stmt)

            if is_superuser_index and row[is_superuser_index]:
                update_stmt = destination_table.update().where(destination_table.c.id == row[0]).values(username='admin')

        destination_connection.commit()


def create_indexes(destination_engine, table_name, column_name):
    destination_metadata = MetaData()
    destination_table = Table(table_name, destination_metadata, autoload_with=destination_engine)

    #r = ReflectedTable(destination_engine)
    #r.create_index(destination_table, "link")
    #r.create_index(destination_table, "title")
    #r.create_index(destination_table, "date_published")


def obfuscate_all(destination_engine):
    obfuscate_user_table("user", destination_engine)

    with destination_engine.connect() as connection:
        r = ReflectedTable(engine=destination_engine, connection=connection)
        r.truncate_table("dataexport")
        r.truncate_table("usersearchhistory")
        r.truncate_table("credentials")


#### SQLite


def get_local_engine(run_info):
    workspace = run_info["workspace"]
    user = run_info["user"]
    database = run_info["database"]
    host = run_info["host"]
    password = run_info["password"]

    # Create the database engine
    SOURCE_DATABASE_URL = f"postgresql://{user}:{password}@{host}/{database}"
    source_engine = create_engine(SOURCE_DATABASE_URL)

    return source_engine


def get_sqlite_engine(run_info):
    workspace = run_info["workspace"]

    file_name = workspace+".db"
    DESTINATION_DATABASE_URL = "sqlite:///" + file_name
    destination_engine = create_engine(DESTINATION_DATABASE_URL)

    return destination_engine


def run_db_copy_backup(run_info):
    workspace = run_info["workspace"]
    tables = run_info["tables"]
    empty = run_info["empty"]

    # Create the database engine
    source_engine = get_local_engine(run_info)

    operating_dir = get_workspace_backup_directory(run_info["format"], workspace)
    operating_dir.mkdir(parents=True, exist_ok=True)
    os.chdir(operating_dir)

    destination_engine = get_sqlite_engine(run_info)

    for table in tables:
        table = table.replace(workspace + "_", "")

        source_table = get_engine_table(workspace, table, source_engine)
        create_destionation_table(table, source_table, destination_engine)

        if not empty:
            copy_table(workspace, table, source_engine, destination_engine, override=True, to_sqlite=True)

    return True


def run_db_copy_restore(run_info):
    workspace = run_info["workspace"]
    tables = run_info["tables"]
    empty = run_info["empty"]
    append = run_info["append"]

    destination_engine = get_local_engine(run_info)

    operating_dir = get_workspace_backup_directory(run_info["format"], workspace)
    os.chdir(operating_dir)

    source_engine = get_sqlite_engine(run_info)

    for table in tables:
        table = table.replace(workspace + "_", "")
        copy_table(workspace, table, source_engine, destination_engine, override=False, to_sqlite=False, commit_every_row=True)

    return True


def run_db_copy_backup_auth(run_info):
    workspace = run_info["workspace"]

    # Create the database engine
    source_engine = get_local_engine(run_info)

    operating_dir = get_workspace_backup_directory(run_info["format"], workspace)
    operating_dir.mkdir(parents=True, exist_ok=True)
    os.chdir(operating_dir)

    destination_engine = get_sqlite_engine(run_info)

    source_table = get_engine_table("auth", "user", source_engine)
    create_destionation_table("user", source_table, destination_engine)

    copy_table("auth", "user", source_engine, destination_engine, override=True, to_sqlite=True)

    return True


def backup_workspace(run_info):
    """
    @note table order is important

    mapping:
      file : tables
    """
    print("--------------------")
    print(run_info["workspace"])
    print("--------------------")

    workspace = run_info["workspace"]

    for table in tables_to_create:
        new_run_info = dict(run_info)
        new_key = "instance_" + table

        new_run_info["tables"] = []
        new_run_info["output_file"] = new_key

        table_name = workspace + "_" + table
        new_run_info["tables"].append(table_name)

        if new_run_info["format"] == "sqlite":
            if table not in tables_to_backup:
                new_run_info["empty"] = True
            if not run_db_copy_backup(new_run_info):
                return False
        else:
            if table in tables_to_backup:
                if not run_pg_dump_backup(new_run_info):
                    return False

    if run_info["format"] == "custom":
        new_run_info = dict(run_info)
        new_run_info["tables"] = ['auth_user']
        new_run_info["workspace"] = 'auth'
        new_run_info["output_file"] = "auth_user"
        if not run_pg_dump_backup(new_run_info):
            return False

    if run_info["format"] == "sqlite":
        run_db_copy_backup_auth(run_info)

        destination_engine = get_sqlite_engine(run_info)

        create_indexes(destination_engine, "linkdatamodel", "link")
        create_indexes(destination_engine, "linkdatamodel", "title")
        create_indexes(destination_engine, "linkdatamodel", "date_published")

        obfuscate_all(destination_engine)

    return True


def restore_workspace(run_info):
    """
    @note table order is important
    """
    workspace = run_info["workspace"]

    print("--------------------")
    print(run_info["workspace"])
    print("--------------------")

    if not run_info["append"]:
        for table in tables_to_backup:
            key = "./instance_" + table
            tables = [workspace + "_" + table]

            run_info["output_file"] = key
            run_info["tables"] = tables

            if not truncate_all(run_info):
                print("Could not truncate table")
                return

    for table in tables_to_backup:
        key = "./instance_" + table
        tables = [workspace + "_" + table]

        new_run_info = dict(run_info)

        new_key = key.replace("instance", workspace)
        new_run_info["output_file"] = new_key
        new_run_info["tables"] = []

        for item in tables:
            table_name = item.replace("instance", workspace)
            new_run_info["tables"].append(table_name)

        if new_run_info["format"] == "sqlite":
            if not run_db_copy_restore(new_run_info):
                return False
        else:
            if not run_pg_restore(new_run_info):
                return False

    reset_tables_index_sequence(tablemapping, run_info)

    return True


def run_sql_for_workspaces(run_info, sql_command):
    print("--------------------")
    print(run_info["workspace"])
    print("--------------------")

    workspace = run_info["workspace"]

    for table in all_tables:
        call_table = workspace + "_" + table
        call_sql_command = sql_command.replace("{table}", call_table)

        if not run_table_sql(run_info, call_table, call_sql_command):
            return False

    return True


def parse_backup_commandline():
    parser = argparse.ArgumentParser(prog="Backup", description="Backup manager. Please provide .pgpass file, and define your password there.")

    parser.add_argument("-b", "--backup", action="store_true", help="Perform a backup")
    parser.add_argument("-r", "--restore", action="store_true", help="Restore from a backup")
    parser.add_argument("-o", "--output-dir", help="Output directory")
    parser.add_argument("-a", "--analyze", action="store_true", help="Analyze the database")
    parser.add_argument("--vacuum", action="store_true", help="Vacuum the database")
    parser.add_argument("--reindex", action="store_true", help="Reindex the database. Useful to detect errors in consistency")
    parser.add_argument("-s", "--sequence-update", action="store_true", help="Updates sequence numbers")
    parser.add_argument("-U", "--user", default="user", help="Username for the database (default: 'user')")
    parser.add_argument("-d", "--database", default="db", help="Database name (default: 'db')")
    parser.add_argument("-p", "--password", default="", help="Password. Necessary for sqlite format")
    parser.add_argument("-w", "--workspaces", help="Workspace for which to perform backup/restore. If not specified - all")
    parser.add_argument("-D", "--debug", help="Enable debug output")  # TODO implement debug
    parser.add_argument("-i", "--ignore-errors", action="store_true", help="Ignore errors during the operation")
    parser.add_argument("--empty", action="store_true", help="Creates empty table version during backup")
    parser.add_argument("--append", action="store_true", help="Appends data during restore, does not clear tables")
    parser.add_argument("-f", "--format", default="custom", choices=["custom", "plain", "sql", "sqlite"],
                        help="Format of the backup (default: 'custom'). Choices: 'custom', 'plain', or 'sql'.")

    parser.add_argument("--host", default="127.0.0.1", help="Host address for the database (default: 127.0.0.1)")

    return parser, parser.parse_args()


class Backup(object):
    def __init__(self, args):
        self.args = args

    def process(self):
        workspaces = []

        if self.args.workspaces:
            if self.args.workspaces.find(",") >= 0:
                workspaces = self.args.workspaces.split(",")
            else:
                workspaces = [self.args.workspaces]
        else:
            print("Provide workspace")
            sys.exit(1)

        if self.args.output_dir:
            output_directory = Path(self.args.output_dir)

        start_time = time.time()

        errors = False

        for workspace in workspaces:
            run_info = {}
            run_info["workspace"] = workspace
            run_info["user"] = self.args.user
            run_info["database"] = self.args.database
            run_info["host"] = self.args.host
            run_info["format"] = self.args.format
            run_info["password"] = self.args.password
            run_info["empty"] = self.args.empty
            run_info["append"] = self.args.append

            if self.args.ignore_errors:
                run_info["ignore_errors"] = True

            if self.args.backup and not backup_workspace(run_info):
                print("Leaving because of errors")
                errors = True
                break

            if self.args.restore and not restore_workspace(run_info):
                print("Leaving because of errors")
                errors = True
                break

            if self.args.analyze and not run_sql_for_workspaces(run_info, "ANALYZE {table};"):
                print("Leaving because of errors")
                errors = True
                break

            if self.args.vacuum and not run_sql_for_workspaces(run_info, "VACUUM {table};"):
                print("Leaving because of errors")
                errors = True
                break

            if self.args.reindex and not run_sql_for_workspaces(run_info, "REINDEX TABLE {table};"):
                print("Leaving because of errors")
                errors = True
                break

            sql_text = "SELECT setval('{table}_id_seq', COALESCE((SELECT MAX(id) FROM {table}), 1));"
            if self.args.sequence_update and not run_sql_for_workspaces(run_info, sql_text):
                print("Leaving because of errors")
                errors = True
                break

        if errors:
            print("There were errors")
        else:
            print("All calls were successful")

        elapsed_time_seconds = time.time() - start_time
        elapsed_minutes = int(elapsed_time_seconds // 60)
        elapsed_seconds = int(elapsed_time_seconds % 60)
        print(f"Time: {elapsed_minutes}:{elapsed_seconds}")


def main():
    parser, args = parse_backup_commandline()

    if not args.backup and not args.restore and not args.analyze and not args.vacuum and not args.reindex and not args.sequence_update:
        parser.print_help()

    else:
        b = Backup(args)
        b.process()


if __name__ == "__main__":
    main()
